﻿//---------------------------------------------------------------------------
//
//   ExportRIB
//
//          Copyright(C) 1999-2015, tetraface Inc.
//
//      Sample of Export plug-in for exporting a RIB (RenderMan) file.
//      Created DLL must be installed in "Plugins\Export" directory.
//
//    　RIB(RenderMan)形式を出力するエクスポートプラグインのサンプル。
//    　作成したDLLは"Plugins\Export"フォルダに入れる必要がある。
//    　RenderManには詳しくないので、このサンプルではあまり多くの情報を出
//　　力していない。
//
//---------------------------------------------------------------------------

#define WIN32_LEAN_AND_MEAN
#define NOMINMAX
#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <vector>
#include "MQPlugin.h"
#include "MQ3DLib.h"
#include "FileLib.h"

BOOL SaveRIB(const char *filename, MQDocument doc);


//---------------------------------------------------------------------------
//  DllMain
//---------------------------------------------------------------------------
BOOL APIENTRY DllMain( HANDLE hModule, 
                       DWORD  ul_reason_for_call, 
                       LPVOID lpReserved
					 )
{
	//プラグインとしては特に必要な処理はないので、何もせずにTRUEを返す
    return TRUE;
}

//---------------------------------------------------------------------------
//  MQGetPlugInID
//    プラグインIDを返す。
//    この関数は起動時に呼び出される。
//---------------------------------------------------------------------------
MQPLUGIN_EXPORT void MQGetPlugInID(DWORD *Product, DWORD *ID)
{
	// プロダクト名(制作者名)とIDを、全部で64bitの値として返す
	// 値は他と重複しないようなランダムなもので良い
	*Product = 0x56A31D20;
	*ID      = 0xC486EA1D;
}

//---------------------------------------------------------------------------
//  MQGetPlugInName
//    プラグイン名を返す。
//    この関数は[プラグインについて]表示時に呼び出される。
//---------------------------------------------------------------------------
MQPLUGIN_EXPORT const char *MQGetPlugInName(void)
{
	// プラグイン名
	return "RIB Exporter   Copyright(C) 1999-2015, tetraface Inc.";
}

//---------------------------------------------------------------------------
//  MQGetPlugInType
//    プラグインのタイプを返す。
//    この関数は起動時に呼び出される。
//---------------------------------------------------------------------------
MQPLUGIN_EXPORT int MQGetPlugInType(void)
{
	// 出力用プラグインである
	return MQPLUGIN_TYPE_EXPORT;
}

//---------------------------------------------------------------------------
//  MQEnumFileType
//    入力または出力可能なファイルタイプを返す。
//    ファイルタイプは別名保存時のダイアログに表示される。
//    この関数は起動時に呼び出される。
//---------------------------------------------------------------------------
MQPLUGIN_EXPORT const char *MQEnumFileType(int index)
{
	switch(index){
	case 0: return "RenderMan (*.rib)";
	}
	return NULL;
}

//---------------------------------------------------------------------------
//  MQEnumFileExt
//    入力または出力可能な拡張子を返す。
//    この関数は起動時に呼び出される。
//---------------------------------------------------------------------------
MQPLUGIN_EXPORT const char *MQEnumFileExt(int index)
{
	switch(index){
	case 0: return "rib";
	}
	return NULL;
}

//---------------------------------------------------------------------------
//  MQExportFile
//    [別名で保存]で保存するときに呼び出される。
//---------------------------------------------------------------------------
MQPLUGIN_EXPORT BOOL MQExportFile(int index, const char *filename, MQDocument doc)
{
	switch(index){
	case 0: return SaveRIB(filename, doc);
	}
	return FALSE;
}

std::string EncodeValue(float value, int digit = 8, int max_digit = 10)
{
	float absval = fabs(value);
	int d = (absval > 0) ? (int)floor(log10(absval)) : 0;

	char fmt[16], buf[64];
	sprintf_s(fmt, "%%0.%df", std::min(std::max(0, digit-d-1), max_digit));
	sprintf_s(buf, fmt, value);

	std::string str(buf);
	size_t pos = str.length();
	bool other = false;
	for(size_t i = str.length(); i > 0; ){
		i--;

		if(str[i] == '.'){
			if(other){
				str = str.substr(0, pos);
				break;
			}else{
				str = str.substr(0, i);
				break;
			}
		}else if(str[i] == '0'){
			if(!other){
				pos = i;
			}
		}else if(str[i] >= '1' && str[i] <= '9'){
			other = true;
		}else{
			break;
		}
	}

	return str;
}

void MatrixMultiply3x3(MQMatrix& dst, const MQMatrix& m1, const MQMatrix& m2)
{
	MQMatrix t1(m1);
	MQMatrix t2(m2);

	dst._11 = t1._11 * t2._11 + t1._12 * t2._21 + t1._13 * t2._31;
	dst._12 = t1._11 * t2._12 + t1._12 * t2._22 + t1._13 * t2._32;
	dst._13 = t1._11 * t2._13 + t1._12 * t2._23 + t1._13 * t2._33;
	dst._21 = t1._21 * t2._11 + t1._22 * t2._21 + t1._23 * t2._31;
	dst._22 = t1._21 * t2._12 + t1._22 * t2._22 + t1._23 * t2._32;
	dst._23 = t1._21 * t2._13 + t1._22 * t2._23 + t1._23 * t2._33;
	dst._31 = t1._31 * t2._11 + t1._32 * t2._21 + t1._33 * t2._31;
	dst._32 = t1._31 * t2._12 + t1._32 * t2._22 + t1._33 * t2._32;
	dst._33 = t1._31 * t2._13 + t1._32 * t2._23 + t1._33 * t2._33;
}

void MatrixRotateZ(MQMatrix& mtx, float radian)
{
	MQMatrix z_rotate;
	z_rotate.Identify();
	z_rotate._11 = z_rotate._22 = cos(radian);
	z_rotate._12 = sin(radian);
	z_rotate._21 = -z_rotate._12;
	MatrixMultiply3x3(mtx, mtx, z_rotate);
}

MQMatrix GetCameraMatrixFromAngle(MQAngle angle)
{
	float ch = cos(RAD(angle.head));
	float sh = sin(RAD(angle.head));
	float cp = cos(RAD(angle.pich));
	float sp = sin(RAD(angle.pich));
	
	MQMatrix mtx;
	mtx.Identify();
	mtx._11 = ch;
	mtx._12 = sh*sp;
	mtx._13 = sh*cp;
	mtx._21 = 0;
	mtx._22 = cp;
	mtx._23 = -sp;
	mtx._31 = sh;
	mtx._32 = -ch*sp;
	mtx._33 = -ch*cp;
	if(angle.bank != 0) MatrixRotateZ(mtx, RAD(angle.bank));
	return mtx;
}

// RIB形式で出力
BOOL SaveRIB(const char *filename, MQDocument doc)
{
	int i, face, mat_index;
	MQMaterial mat;
	MQFileDialogInfo dlginfo;
	char path[MAX_PATH], name[64];

	// 座標軸変換用ダイアログの表示
	memset(&dlginfo, 0, sizeof(dlginfo));
	dlginfo.dwSize = sizeof(dlginfo);
	dlginfo.axis_x = MQFILE_TYPE_RIGHT;
	dlginfo.axis_y = MQFILE_TYPE_UP;
	dlginfo.axis_z = MQFILE_TYPE_BACK;
	dlginfo.softname = "RenderMan";
	MQ_ShowFileDialog("RIB Export", &dlginfo);

	// ファイルを作成・オープンする
	FILE *fh;
	errno_t ret_s = fopen_s(&fh, filename, "w");
	if(ret_s != 0)
		return FALSE;

	MQScene scene = doc->GetScene(0);

	// 複数のオブジェクトを１つに合成するために、作業用オブジェクトを作成する
	MQObject obj = MQ_CreateObject();

	// 作業用オブジェクトにすべてのオブジェクトを合成
	int objcount = doc->GetObjectCount();
	for(i=0; i<objcount; i++)
	{
		MQObject thisobj = doc->GetObject(i);
		if(thisobj == NULL) 
			continue;
		// フリーズなどにより内容が変更されてしまうので、オブジェクトを複製しておく
		thisobj = thisobj->Clone();
		// 曲面などの属性をフリーズ
		thisobj->Freeze(MQOBJECT_FREEZE_ALL);
		// 作業用オブジェクトに合成する
		obj->Merge(thisobj);
		// 複製したオブジェクトはもう不要なので破棄
		thisobj->DeleteThis();
	}

	// 面と頂点の数を取得
	int face_count = obj->GetFaceCount();
	int vert_count = obj->GetVertexCount();

	// 座標軸をダイアログで指定したものに変換
	for(i=0; i<vert_count; i++){
		if(obj->GetVertexRefCount(i) == 0) continue;
		MQPoint p = obj->GetVertex(i);
		MQ_ExportAxis(&dlginfo, &p, 1);
		obj->SetVertex(i, p);
	}

	MQObjNormal normal(obj);

	// ファイルの先頭にデータを書き込み
	fprintf(fh, "##RenderMan RIB-Structure 1.1\n");
	fprintf(fh, "version 3.03\n");
	fprintf(fh, "\n");

	fprintf(fh, "Option \"searchpath\" \"shader\" [\".:&\"]\n");
	GetFileNameFromPath(path, filename);
	fprintf(fh, "Display \"%s.tif\" \"file\" \"rgba\"\n", path);
	float scene_scale = 1.0f;
	if(scene->GetOrtho()){
		scene_scale = 0.001f * scene->GetZoom();
		fprintf(fh, "Projection \"orthographic\" \n");
	}else{
		fprintf(fh, "Projection \"perspective\" \"fov\" %.1f\n", scene->GetFOV()/3.14159f*180.0f);
	}
	MQAngle angle = scene->GetCameraAngle();
	MQMatrix cammtx = GetCameraMatrixFromAngle(angle);
	MQPoint campos = scene->GetCameraPosition() * cammtx * scene_scale;
	fprintf(fh, "Translate %f %f %f\n", -campos.x, -campos.y, -campos.z);
	fprintf(fh, "Rotate %f 0 0 1\n", angle.bank);
	fprintf(fh, "Rotate %f 1 0 0\n", -angle.pitch);
	fprintf(fh, "Rotate %f 0 1 0\n", -angle.head);
	if(scene_scale != 1.0f){
		fprintf(fh, "Scale %f %f %f\n", scene_scale, scene_scale, scene_scale);
	}
	fprintf(fh, "\n");

	MQColor amb_color = scene->GetGlobalAmbientColor();
	fprintf(fh, "WorldBegin\n");
	fprintf(fh, "\n");
	fprintf(fh, "LightSource \"ambientlight\" 1 \"lightcolor\" [%.3f %.3f %.3f]\n", amb_color.r, amb_color.g, amb_color.b);
	fprintf(fh, "\n");
	fprintf(fh, "Declare \"shadows\" \"string\"\n");
	fprintf(fh, "Attribute \"light\" \"shadows\" \"on\"\n");
	int light_num = scene->GetGlobalDirectionalLightNum();
	for(i=0; i<light_num; i++){
		MQPoint dir = scene->GetGlobalDirectionalLightDirection(i);
		MQ_ExportAxis(&dlginfo, &dir, 1);
		dir.normalize();
		MQColor col = scene->GetGlobalDirectionalLightColor(i);
		fprintf(fh, "LightSource \"distantlight\" %d \"from\" [%f %f %f] \"to\" [0 0 0] \"lightcolor\" [%.3f %.3f %.3f]\n", 
			i+2, -dir.x, -dir.y, -dir.z, col.r, col.g, col.b);
	}
	fprintf(fh, "\n");

	// それぞれの材質がいくつの面に使われているかどうかを調べておく
	int mat_count = doc->GetMaterialCount();
	int *mat_used = (int *)malloc(sizeof(int) * (mat_count+1));
	for(i=0; i<=mat_count; i++)
		mat_used[i] = 0;
	for(face=0; face<face_count; face++)
	{
		mat_index = obj->GetFaceMaterial(face);
		mat_used[(mat_index>=0) ? mat_index : mat_count]++; // 未着色面の場合は配列の最後にデータをセット
	}

	// 材質ごとにAttributeとしてポリゴン情報を記録する
	for(mat_index=0; mat_index<=mat_count; mat_index++)
	{
		int mi;

		// この材質が使われていなければパス
		if(mat_used[mat_index] == 0)
			continue;

		if(mat_index == mat_count) {
			mat = MQ_CreateMaterial(); // 未着色面の場合は一時的に材質を作成
			mi = -1;
		} else {
			mat = doc->GetMaterial(mat_index);
			mi = mat_index;
		}
	
		fprintf(fh, "AttributeBegin\n");

		// コメントとして材質名を記録しておく
		mat->GetName(name, sizeof(name));
		fprintf(fh, "  # %s\n", name);

		MQColor color = mat->GetColor();
		float alpha = mat->GetAlpha();
		fprintf(fh, "  Color [ %.3f %.3f %.3f ]\n", color.r, color.g, color.b);
		fprintf(fh, "  Opacity [ %.3f %.3f %.3f ]\n", alpha, alpha, alpha);

		char texname[MAX_PATH];
		mat->GetTextureName(texname, MAX_PATH);

		switch(mat->GetShader()){
		case MQMATERIAL_SHADER_CONSTANT:
			fprintf(fh, "  Surface \"constant\"\n");
			break;
		case MQMATERIAL_SHADER_LAMBERT:
			fprintf(fh, "  Surface \"matte\" \"Ka\" %.3f \"Kd\" %.3f\n", mat->GetAmbient(), mat->GetDiffuse());
			break;
		default:
			{
				MQColor spc_col = mat->GetSpecularColor();
				if(texname[0] != '\0'){
					fprintf(fh, "  Surface \"paintedplastic\" \"Ka\" %.3f \"Kd\" %.3f \"Ks\" 1 \"roughness\" %.3f \"specularcolor\" [%.3f %.3f %.3f] \"texturename\" \"%s\"\n", 
						mat->GetAmbient(), mat->GetDiffuse(), mat->GetPower(), spc_col.r, spc_col.g, spc_col.b, texname);
				}else{
					fprintf(fh, "  Surface \"plastic\" \"Ka\" %.3f \"Kd\" %.3f \"Ks\" 1 \"roughness\" %.3f \"specularcolor\" [%.3f %.3f %.3f]\n", 
						mat->GetAmbient(), mat->GetDiffuse(), mat->GetPower(), spc_col.r, spc_col.g, spc_col.b);
				}
			}
			break;
		}

		for(face=0; face<face_count; face++)
		{
			// この面に割り当てられた材質を調べる
			if(obj->GetFaceMaterial(face) == mi)
			{
				int vert, count;

				// 面中の頂点数と頂点インデックスを取得
				count = obj->GetFacePointCount(face);

				std::vector<int> vert_index(count);
				obj->GetFacePointArray(face, &(*vert_index.begin()));

				fprintf(fh, "  Polygon \"P\" [ ");
				for(vert=0; vert<count; vert++){
					MQPoint p = obj->GetVertex(vert_index[vert]);
					fprintf(fh, "%s %s %s ", EncodeValue(p.x).c_str(), EncodeValue(p.y).c_str(), EncodeValue(p.z).c_str());
				}
				fprintf(fh, "] \"N\" [ ");
				for(vert=0; vert<count; vert++){
					MQPoint p = normal.Get(face, vert);
					fprintf(fh, "%s %s %s ", EncodeValue(p.x).c_str(), EncodeValue(p.y).c_str(), EncodeValue(p.z).c_str());
				}
				if(texname[0] != '\0'){
					fprintf(fh, "] \"st\" [ ");
					std::vector<MQCoordinate> uvarray(count);
					obj->GetFaceCoordinateArray(face, &(*uvarray.begin()));
					for(vert=0; vert<count; vert++){
						fprintf(fh, "%.5f %.5f ", uvarray[vert].u, uvarray[vert].v);
					}
				}
				fprintf(fh, "]\n");
			}
		}

		fprintf(fh, "AttributeEnd\n");
		fprintf(fh, "\n");

		// 未着色面用に一時的に作った材質を削除
		if(mat_index == mat_count)
			mat->DeleteThis();
	}

	// ファイルの終端のデータを書き込み
	fprintf(fh, "WorldEnd\n");
	fprintf(fh, "\n");

	// ファイルをクローズ
	fclose(fh);

	// 不要になったデータを破棄
	free(mat_used);
	obj->DeleteThis();

	return TRUE;
}

